#include "cWeapon.h"
#include "cObject.h"

#include "macros.h"

#include <GL/gl.h>

#define DETAIL 0


//--------------------------------


cWeapon::cWeapon()
{
    //printf("cWeapon::cWeapon()\n");
    mMountfv = mBasefv = NULL;
    mOwner = NULL;
    mReloading = 0.0f;
}


cWeapon::cWeapon(cWeapon *source)
{
    //printf("cWeapon::cWeapon(cObject*)\n");
    mBasefv    = source->mBasefv;
    mMountfv   = source->mMountfv;
    mOwner     = source->mOwner;
    mReloading = source->mReloading;
}


int cWeapon::damageAllByCylinder(float* position, float* velocity, float spf, float radius, int roles, float damage)
{
    int damaged = 0;
    
    float *v = velocity;
    float *a = position;
    float b[] = { a[0] - v[0] * spf, a[1] - v[1] * spf, a[2] - v[2] * spf };
    //printf("b' %.2f %.2f %.2f\n", b[0],b[1],b[2]);
    
    float length = sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
    //printf("length %.6f\n", length);

    std::list<cObject*>* role = cWorld::instance->filterByRole(mOwner, roles, true, -1);
    if (role) {
        //printf("roles: \n"); cWorld::instance->printNames(role);
        //printf("roles %i\n", role->size());
        std::list<cObject*>* range = cWorld::instance->filterByRange(mOwner, a, 0.0f, length, -1, role);
        if (range) {
            //printf("range: \n"); cWorld::instance->printNames(range);
            //printf("range %i\n", range->size());

                looplist(i, cObject*, *range) {
                    // todo: min dist
                    cObject::rImpact* imp = ((cObject*)(*i))->damageByBeam(a, b, radius, damage, true);
                    if (imp != NULL) {
                        damaged++;
                        delete imp;
                    }
                }

//            std::list<cObject*>* line = cWorld::instance->filterByBeam(mOwner, a, b, radius*5, -1, range);
//            if (line) {
//                //printf("line: \n"); cWorld::instance->printNames(line);
//                printf("line %i\n", line->size());
//                looplist(i, cObject*, *line) {
//                    // todo: min dist
//                    cObject::rImpact* imp = ((cObject*)(*i))->damageByBeam(a, b, radius, damage, true);
//                    if (imp != NULL) {
//                        damaged++;
//                        delete imp;
//                    }
//                }
//                // todo: damage nearest target, if not really hit then find next nearest ...
//                // Sort them?
//                // Binary search through filterByRange?
//                delete line;
//            }
            delete range;
        }
        delete role;
    }
    return damaged;
}

//--------------------------------


cMachineGun::cMachineGun()
{
}


void cMachineGun::fire(cObject *target)
{
    if (!ready()) return;
    mReloading = 0.5;
    
    float source[16];
    glPushMatrix();
        glLoadIdentity();
            glMultMatrixf(mBasefv);
                if (mMountfv) glMultMatrixf(mMountfv);
                glGetFloatv(GL_MODELVIEW_MATRIX, source);
    glPopMatrix();
    
    {
        Shrapnel* s = new Shrapnel();
        s->cnt = 4;

        s->pos[0] = source[12];
        s->pos[1] = source[13];
        s->pos[2] = source[14];
        
        float nrm[3];
        float pos2[] = { 0, -1, 0 };
        applyMatrix2(source, pos2);
        vectorsub(nrm, pos2, s->pos);

        scalarmul(s->vel, nrm, 1.4);
        const float variation = 0.01;
        for (int i=0; i<3; i++) s->vel[i] += (50-(rand()%100)) * variation;
        castoff.push_back(s);
    }
    
    
    {
        Shrapnel* s = new Shrapnel();
        s->cnt = 4;

        s->pos[0] = source[12];
        s->pos[1] = source[13];
        s->pos[2] = source[14];

        float nrm[3];
        float pos2[] = { 0, 0, -1 };
        applyMatrix2(source, pos2);
        vectorsub(nrm, pos2, s->pos);

        scalarmul(s->vel, nrm, 200);
        const float variation = 0.03;
        for (int i=0; i<3; i++) s->vel[i] += (50-(rand()%100)) * variation;
        shrapnels.push_back(s);

        float len1 = 0.5;
        float len2 = 200;
        float a[] = {
            s->pos[0]+nrm[0]*len1,
            s->pos[1]+nrm[1]*len1,
            s->pos[2]+nrm[2]*len1
        };
        float b[] = {
            s->pos[0]+nrm[0]*len2,
            s->pos[1]+nrm[1]*len2,
            s->pos[2]+nrm[2]*len2
        };
        
        //printf("a %.2f %.2f %.2f\n", a[0],a[1],a[2]);
        //printf("b %.2f %.2f %.2f\n", b[0],b[1],b[2]);

        const float cylinder_radius = 2.5f;
        const float damage = 10.0f;
        float v[] = { -(b[0]-a[0]), -(b[1]-a[1]), -(b[2]-a[2]) };
        //float b[] = { a[0] - v[0] * spf, a[1] - v[1] * spf, a[2] - v[2] * spf };
        //printf("v %.2f %.2f %.2f\n", v[0],v[1],v[2]);
        int damaged = this->damageAllByCylinder(a, v, 10.0f, cylinder_radius, (1UL<<cObject::COLLIDEABLE)|(1UL<<cObject::DAMAGEABLE), damage);
        //printf("Damaged: %i\n", damaged);
    }
}


void cMachineGun::animate(float spf)
{
    for (std::list<Shrapnel*>::iterator i = shrapnels.begin(); i != shrapnels.end();) {
        Shrapnel* s = *i++;
        s->pos[0] += s->vel[0] * spf;
        s->pos[1] += s->vel[1] * spf;
        s->pos[2] += s->vel[2] * spf;
        s->cnt -= spf;
        if (s->cnt < 0) {
            shrapnels.remove(s);
            delete s;
        }
    }
    for (std::list<Shrapnel*>::iterator i = castoff.begin(); i != castoff.end();) {
        Shrapnel* s = *i++;
        s->pos[0] += s->vel[0] * spf;
        s->pos[1] += s->vel[1] * spf;
        s->pos[2] += s->vel[2] * spf;
        s->cnt -= spf;
        if (s->cnt < 0) {
            castoff.remove(s);
            delete s;
        }
    }
    mReloading -= spf;
    if (mReloading < 0) mReloading = 0.0f;
};


void cMachineGun::drawEffect()
{
    glPushMatrix();
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glLineWidth(2);
    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glDepthMask(GL_FALSE);

    for (std::list<Shrapnel*>::iterator i = shrapnels.begin(); i != shrapnels.end();) {
        Shrapnel* s = *i++;
        const float len1 = 0.01;
        const float len2 = 0.02;
        glBegin(GL_LINE_STRIP);
            glColor4f(1,1,1,1);
            glVertex3fv(s->pos);
            glColor4f(1,1,0,1);
            glVertex3f(s->pos[0] - s->vel[0]*len1, s->pos[1] - s->vel[1]*len1, s->pos[2] - s->vel[2]*len1);
            glColor4f(1,0,0,0.3);
            glVertex3f(s->pos[0] - s->vel[0]*len2, s->pos[1] - s->vel[1]*len2, s->pos[2] - s->vel[2]*len2);
        glEnd();
    }

    for (std::list<Shrapnel*>::iterator i = castoff.begin(); i != castoff.end();) {
        Shrapnel* s = *i++;
        const float len1 = 0.01;
        const float len2 = 0.02;
        glBegin(GL_LINE_STRIP);
            glColor4f(1,1,1,1);
            glVertex3fv(s->pos);
            glColor4f(1,1,0,1);
            glVertex3f(s->pos[0] - s->vel[0]*len1, s->pos[1] - s->vel[1]*len1, s->pos[2] - s->vel[2]*len1);
            glColor4f(1,0,0,0.3);
            glVertex3f(s->pos[0] - s->vel[0]*len2, s->pos[1] - s->vel[1]*len2, s->pos[2] - s->vel[2]*len2);
        glEnd();
    }

    glPopMatrix();
    glPopAttrib();
}


void cMachineGun::drawHUD() {
    float a = 0.9;
    float b = 0.3;
    glBegin(GL_LINES);
        glColor4f(1,1,0,b);
        glVertex3f(0.25, 0.5, 0.0);
        glColor4f(1,1,1,a);
        glVertex3f(0.35, 0.5, 0.0);
        glColor4f(1,1,0,b);
        glVertex3f(0.45, 0.5, 0.0);
        glColor4f(1,1,1,a);
        glVertex3f(0.55, 0.5, 0.0);
        glColor4f(1,1,0,b);
        glVertex3f(0.65, 0.5, 0.0);
        glColor4f(1,1,1,a);
        glVertex3f(0.75, 0.5, 0.0);
    glEnd();
}


//--------------------------------


cPlasmaGun::cPlasmaGun()
{
}


void cPlasmaGun::fire(cObject *target)
{
    if (!ready()) return;
    mReloading = 0.9;
    
    float source[16];
    glPushMatrix();
        glLoadIdentity();
            glMultMatrixf(mBasefv);
                if (mMountfv) glMultMatrixf(mMountfv);
                glGetFloatv(GL_MODELVIEW_MATRIX, source);
    glPopMatrix();
    
    Shrapnel* s = new Shrapnel();
    s->cnt = 3;
    
    s->pos[0] = source[12];
    s->pos[1] = source[13];
    s->pos[2] = source[14];

    float nrm[3];
    float pos2[] = { 0, 0, -1 };
    applyMatrix2(source, pos2);
    vectorsub(nrm, pos2, s->pos);
    
    scalarmul(s->vel, nrm, 50);
    shrapnels.push_back(s);
}


void cPlasmaGun::animate(float spf)
{
    for (std::list<Shrapnel*>::iterator i = shrapnels.begin(); i != shrapnels.end();) {
        Shrapnel* s = *i++;
        s->pos[0] += s->vel[0] * spf;
        s->pos[1] += s->vel[1] * spf;
        s->pos[2] += s->vel[2] * spf;
        s->cnt -= spf;
        if (s->cnt < 0) {
            shrapnels.remove(s);
            delete s;
        } else {
            int damaged = this->damageAllByCylinder(s->pos, s->vel, spf, 10.0f, (1UL<<cObject::COLLIDEABLE)|(1UL<<cObject::DAMAGEABLE), 10);
            if (damaged) {
                shrapnels.remove(s);
                delete s;
            }
        }
    }
    mReloading -= spf;
    if (mReloading < 0) mReloading = 0.0f;
};


void cPlasmaGun::drawEffect()
{
    glPushMatrix();
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT | GL_ALL_ATTRIB_BITS);

    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glDepthMask(GL_FALSE);

    float n[16];
    glGetTransposeInverseRotationMatrix(n);

    for (std::list<Shrapnel*>::iterator i = shrapnels.begin(); i != shrapnels.end();) {
        Shrapnel* s = *i++;
        glPushMatrix();
            glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glMultMatrixf(n);
                glColor4f(0.1, 0.1, 0.6, 0.99f);
                glDisk(9, 1.9*0.1f);
                glColor4f(0.8, 0.8, 0.1, 0.6f);
                glDisk(7, 1.9*0.07f);
        glPopMatrix();
    }

    glPopMatrix();
    glPopAttrib();
}


void cPlasmaGun::drawHUD() {
    float a = 0.9;
    float b = 0.3;
    glBegin(GL_LINES);
        glColor4f(0,0,1,b);
        glVertex3f(0.25, 0.5, 0.0);
        glColor4f(1,1,1,a);
        glVertex3f(0.75, 0.5, 0.0);
    glEnd();
}


//--------------------------------


cSparkGun::cSparkGun()
{
}


void cSparkGun::fire(cObject *target)
{
    if (!ready()) return;
    mReloading = 0.8;
    
    float source[16];
    glPushMatrix();
        glLoadIdentity();
            glMultMatrixf(mBasefv);
                if (mMountfv) glMultMatrixf(mMountfv);
                glGetFloatv(GL_MODELVIEW_MATRIX, source);
    glPopMatrix();
    
    Shrapnel* s = new Shrapnel();
    s->cnt = 1;
    
    s->pos[0] = source[12];
    s->pos[1] = source[13];
    s->pos[2] = source[14];

    float nrm[3];
    float pos2[] = { 0, 0, -1 };
    applyMatrix2(source, pos2);
    vectorsub(nrm, pos2, s->pos);
    
    scalarmul(s->vel, nrm, 30);
    shrapnels.push_back(s);
}


void cSparkGun::animate(float spf)
{
    for (std::list<Shrapnel*>::iterator i = shrapnels.begin(); i != shrapnels.end();) {
        Shrapnel* s = *i++;
        s->pos[0] += s->vel[0] * spf;
        s->pos[1] += s->vel[1] * spf;
        s->pos[2] += s->vel[2] * spf;
        s->cnt -= spf;
        if (s->cnt < 0) {
            shrapnels.remove(s);
            delete s;
        } else {
            int damaged = this->damageAllByCylinder(s->pos, s->vel, spf, 10.0f, (1UL<<cObject::COLLIDEABLE)|(1UL<<cObject::DAMAGEABLE), 10);
            if (damaged) {
                shrapnels.remove(s);
                delete s;
            }
        }
    }
    mReloading -= spf;
    if (mReloading < 0) mReloading = 0.0f;
};


void cSparkGun::drawEffect()
{
    glPushMatrix();
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glDepthMask(GL_FALSE);

    float n[16];
    glGetTransposeInverseRotationMatrix(n);

    for (std::list<Shrapnel*>::iterator i = shrapnels.begin(); i != shrapnels.end();) {
        Shrapnel* s = *i++;
        glPushMatrix();
            glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glMultMatrixf(n);
                glColor4f(0.6, 0.1, 0.1, 0.99f);
                glDisk(9+DETAIL, 5*0.1f+0.006*(rand()%100));
                glColor4f(0.9, 0.9, 0.4, 0.99f);
                glDisk(7+DETAIL, 5*0.07f+0.006*(rand()%100));
        glPopMatrix();
    }

    glPopMatrix();
    glPopAttrib();
}


//--------------------------------

cRaybeam::cRaybeam() : raylife(0)
{
};


void cRaybeam::fire(cObject *target)
{
    if (!ready()) return;
    mReloading = 2.0;
    raylife = 0.7;
    //mReloading = raylife;
}


void cRaybeam::animate(float spf) {
    raylife -= spf;
    if (raylife < 0) raylife = 0.0f;
    mReloading -= spf;
    if (mReloading < 0) mReloading = 0.0f;
}


void cRaybeam::drawSolid() {
};


void cRaybeam::drawEffect() {
    if (raylife == 0) return;
    
    const float len = 10;
    struct appearance {
        float width, length;
        float icolor[4];
        float ocolor[4];
    } app[] = {
        0.10,   8*len,  {0.3, 0.3, 0.3, 1},  {0.3, 0.0, 0.0, 1},
        0.40,   8*len,  {0.5, 0.5, 0.5, 1},  {0.3, 0.0, 0.0, 1},
        0.80,   8*len,  {0.9, 0.9, 0.9, 1},  {0.3, 0.0, 0.0, 1},
        0.85,   8*len,  {1.0, 1.0, 1.0, 1},  {0.4, 0.0, 0.0, 1},
        0.90, 8.2*len,  {1.0, 1.0, 1.0, 1},  {0.7, 0.0, 0.0, 1},
        0.60, 8.4*len,  {0.9, 0.9, 0.9, 1},  {0.3, 0.0, 0.0, 1},
        0.30, 8.5*len,  {0.7, 0.7, 0.7, 1},  {0.3, 0.0, 0.0, 1},
    };

    int lifeindex = (int) ((0.7-raylife) * 10);
    if (lifeindex > sizeof(app)/sizeof(appearance) - 1) lifeindex = sizeof(app)/sizeof(appearance) - 1;
    if (lifeindex < 0) lifeindex = 0;
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);
        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        glDepthMask(GL_FALSE);
        
        glPushMatrix();
            glMultMatrixf(this->mBasefv);
            if (mMountfv) glMultMatrixf(this->mMountfv);


            /*
            float m[16];
            glGetFloatv(GL_MODELVIEW_MATRIX, m);
            glPushMatrix();
            glLoadIdentity();
            float right[] = {1,0,0};
            gluLookAt(0,0,0, m[12],m[13],m[14], m[8],m[9],m[10]);
            //gluLookAt(0,0,0, m[12],m[13],m[14], 0,1,0);
            memcpy(right, &m[4], sizeof(float)*3);
            glPopMatrix();
            */
            
            float m[16];
            glGetFloatv(GL_MODELVIEW_MATRIX, m);
            float* mpnt = &m[12]; // Vector that points to the Mountpoint relative to the eye.
            //printVector(mpnt);
            float* mfwd = &m[8]; // Mountpoint-Forward Vector relative to the eye.
            //printVector(mfwd);
            
            float right[3];
            cross(right, mpnt, mfwd);
            normalize(right, right);
            //printVector(right);
            
            //if (mpnt[2] > 0 || mfwd[2] < 0) right[1] = -right[1]; // Correction
            //if (mpnt[1] > 0 || mfwd[1] < 0) right[1] = -right[1]; // Correction
            
            
            const float w = app[lifeindex].width;
            right[0] *= -w;
            right[1] *= -w;
            right[2] *= -w;
            glBegin(GL_TRIANGLE_STRIP);
                glColor4fv(app[lifeindex].ocolor);
                glVertex3f(-right[0], -right[1], -right[2]-2*w); glVertex3f(-right[0], -right[1], -right[2]-app[lifeindex].length);
                glColor4fv(app[lifeindex].icolor);
                glVertex3f(0, 0.0, 0); glVertex3f(0, 0.0, -app[lifeindex].length);
                glColor4fv(app[lifeindex].ocolor);
                glVertex3f(+right[0], +right[1], +right[2]-2*w); glVertex3f(+right[0], +right[1], +right[2]-app[lifeindex].length);
            glEnd();
            
        glPopMatrix();
    glPopAttrib();
};


void cRaybeam::drawHUD() {
    float a = 0.9;
    float b = 0.6;
    glBegin(GL_LINES);
        glColor4f(1,0.3,0.3,b);
        glVertex3f(0.25, 0.5, 0.0);
        glColor4f(1,0.6,0.6,a);
        glVertex3f(0.75, 0.5, 0.0);
    glEnd();
}

//--------------------------------


cHomingMissile::cHomingMissile()
{
}


void cHomingMissile::fire(cObject *target)
{
    if (!ready()) return;
    mReloading = 0.8;
    
    float source[16];
    glPushMatrix();
        glLoadIdentity();
            glMultMatrixf(mBasefv);
                if (mMountfv) glMultMatrixf(mMountfv);
                glGetFloatv(GL_MODELVIEW_MATRIX, source);
    glPopMatrix();
    
    Missile* s = new Missile();
    s->target = target;
    s->fuel = 4;
    s->nextsmoke = 0.0f;
    s->timer = 0.0f;
    
    s->pos[0] = source[12];
    s->pos[1] = source[13];
    s->pos[2] = source[14];
    
    float nrm[3];
    float pos2[] = { 0, 0, -1 };
    
    // Variation of direction.
    //pos2[0] += (100-(rand()%200)) * 0.005;
    //pos2[1] += (100-(rand()%200)) * 0.005 + 0.1;
    //pos2[2] += (100-(rand()%200)) * 0.01;

    applyMatrix2(source, pos2);
    vectorsub(nrm, pos2, s->pos);
    
    scalarmul(s->vel, nrm, 1);
    missiles.push_back(s);
}


void cHomingMissile::animate(float spf)
{
    for (std::list<Missile*>::iterator i = missiles.begin(); i != missiles.end();) {
        Missile* s = *i++;

        // Update Smoke from Missile Trails.
        bool removed_smoke = false;
        for (std::list<Smoke*>::iterator j = s->trail.begin(); j != s->trail.end();) {
            Smoke* smoke = *j++;
            smoke->cnt -= spf;
            smoke->timer += spf;
            if (smoke->cnt < 0) {
                s->trail.remove(smoke);
                removed_smoke = true;
            }
        }

        const int MAX_SMOKE = 300;
        // Add new Smoke-Particle whenever an old Smoke-Particel disappears
        // or at specified intervals.
        s->nextsmoke -= spf;
        if (s->nextsmoke <= 0 || removed_smoke) {
            s->nextsmoke = 0.04;
            Smoke* smoke = new Smoke();
            memcpy(smoke->pos, s->pos, sizeof(float)*3);
            smoke->cnt = 0.9f;
            smoke->timer = 0.0f;
            s->trail.push_back(smoke);
        } 

        // Now update the Missile itself
        float tgt[] = { 0,0,0 };
        if (s->target) memcpy(tgt, s->target->mPos, sizeof(float)*3);
        float dir[] = { tgt[0]-s->pos[0], tgt[1]-s->pos[1], tgt[2]-s->pos[2] };
        normalize(dir,dir);
        normalize(s->vel,s->vel);
        const float alpha = 0.96;
        loop3i(s->vel[i] = (s->vel[i]*alpha+dir[i]*(1-alpha)));
        normalize(s->vel,s->vel);
        const float speed = 30;
        s->pos[0] += speed * s->vel[0] * spf;
        s->pos[1] += speed * s->vel[1] * spf;
        s->pos[2] += speed * s->vel[2] * spf;
        s->fuel -= spf;
        s->timer += spf;

        // When fuel is empty just let the Missile and all
        // it's Smoke-Particles disappear.
        if (s->fuel < 0) {
            for (std::list<Smoke*>::iterator j = s->trail.begin(); j != s->trail.end();) {
                Smoke* smoke = *j++;
                s->trail.remove(smoke);
            }
            missiles.remove(s);
            delete s;
        } else {
            float v[] = {speed*s->vel[0], speed*s->vel[1], speed*s->vel[2] };
            int damaged = this->damageAllByCylinder(s->pos, v, spf, 5.0f, (1UL<<cObject::COLLIDEABLE)|(1UL<<cObject::DAMAGEABLE), 10);
            if (damaged) {
                missiles.remove(s);
                delete s;
            }
        }
    }
    mReloading -= spf;
    if (mReloading < 0) mReloading = 0.0f;
};


void cHomingMissile::drawSolid()
{
    glPushMatrix();
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);

    glColor4f(0.7,0.7,0.7,1.0);

    const float size = 0.8;
    const float length = size * 1.4;
    const float radius = size * 0.2;
    const float s1 = radius * sin(  0*PI_OVER_180);
    const float c1 = radius * cos(  0*PI_OVER_180);
    const float s2 = radius * sin(120*PI_OVER_180);
    const float c2 = radius * cos(120*PI_OVER_180);
    const float s3 = radius * sin(240*PI_OVER_180);
    const float c3 = radius * cos(240*PI_OVER_180);
    
    float m[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, m);
    
    for (std::list<Missile*>::iterator i = missiles.begin(); i != missiles.end();) {
        Missile* s = *i++;
        
        float r[3];
        cross(r, s->vel, (&m[4]));
        float u[3];
        cross(u, s->vel, r);
        float b[16];
        memset(b, 0, sizeof(float)*16);
        b[15] = 1;
        memcpy(&b[0], r, sizeof(float)*3);
        memcpy(&b[4], u, sizeof(float)*3);
        memcpy(&b[8], s->vel, sizeof(float)*3);
        //printMatrix(b);
        
        glPushMatrix();
            glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glMultMatrixf(b);
                //glNormAxis();
                glBegin(GL_TRIANGLE_FAN);
                glVertex3f(0,0,length);
                glVertex3f(s1,c1,0);
                glVertex3f(s2,c2,0);
                glVertex3f(s3,c3,0);
                glVertex3f(s1,c1,0);
                glEnd();
        glPopMatrix();
    }

    glPopMatrix();
    glPopAttrib();
}


void cHomingMissile::drawEffect()
{
    //return;
    
    glPushMatrix();
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glDepthMask(GL_FALSE);

    float n[16];
    glGetTransposeInverseRotationMatrix(n);
    
    // Draw Missile's flare.
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    for (std::list<Missile*>::iterator i = missiles.begin(); i != missiles.end();) {
        Missile* s = *i++;
        
        glPushMatrix();
            glTranslatef(s->pos[0], s->pos[1], s->pos[2]);            
                glMultMatrixf(n);
                float size = 2.5 + 2.5 * 0.01f*(rand()%100);
                glColor4f(0.8, 0.5, 0.1, 0.99f);
                glDisk(9+DETAIL, size*0.1f);
                size = 2.5 + 2.5 * 0.01f*(rand()%100);
                glColor4f(0.9, 0.9, 0.9, 0.99f);
                glDisk(7+DETAIL, size*0.07f);
        glPopMatrix();
   }
   
   // Draw Missile's trails.
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   for (std::list<Missile*>::iterator i = missiles.begin(); i != missiles.end();) {
        Missile* s = *i++;
        
        for (std::list<Smoke*>::iterator j = s->trail.begin(); j != s->trail.end();) {
            Smoke* smoke = *j++;
            glPushMatrix();
                glTranslatef(smoke->pos[0], smoke->pos[1], smoke->pos[2]);            
                    glMultMatrixf(n);
                    float m = 2*smoke->cnt;
                    glColor4f(m, m*0.8, m*0.6, 0.05+smoke->cnt);
                    float size = 0.1 + 0.2 * pow(1.0f+smoke->timer, 2.0);
                    glDisk(5+DETAIL, size);
            glPopMatrix();
        }
   }

    glPopMatrix();
    glPopAttrib();
}


//--------------------------------


cExplosion::cExplosion()
{
}


void cExplosion::fire(cObject *target)
{
    if (!ready()) return;
    mReloading = 3.8;
    
    // Shoot several shrapnels around.
    for (int i = 0; i < 15; i++) {
        
        float source[16];
        glPushMatrix();
            glLoadIdentity();
                if (mBasefv) glMultMatrixf(mBasefv);
                else if (mOwner) glTranslatef(mOwner->mPos[0], mOwner->mPos[1], mOwner->mPos[2]);
                    if (mMountfv) glMultMatrixf(mMountfv);
                    glGetFloatv(GL_MODELVIEW_MATRIX, source);
        glPopMatrix();

        Missile* s = new Missile();
        s->target = target;
        s->fuel = 4*2;
        s->nextsmoke = 0.0f;
        s->timer = 0.0f;

        s->pos[0] = source[12];
        s->pos[1] = source[13];
        s->pos[2] = source[14];

        float nrm[3];
        float pos2[] = { 0, 0, -1 };

        pos2[0] = (100-(rand()%200)) * 0.01;
        pos2[1] = 0.4+(100-(rand()%200)) * 0.01;
        pos2[2] = (100-(rand()%200)) * 0.01;

        applyMatrix2(source, pos2);
        vectorsub(nrm, pos2, s->pos);

        scalarmul(s->vel, nrm, 1);
        missiles.push_back(s);
    } // for
}


void cExplosion::animate(float spf)
{
    for (std::list<Missile*>::iterator i = missiles.begin(); i != missiles.end();) {
        Missile* s = *i++;

        // Update Smoke from Missile Trails.
        bool removed_smoke = false;
        for (std::list<Smoke*>::iterator j = s->trail.begin(); j != s->trail.end();) {
            Smoke* smoke = *j++;
            smoke->cnt -= spf;
            smoke->timer += spf;
            if (smoke->cnt < 0) {
                s->trail.remove(smoke);
                removed_smoke = true;
            }
        }

        const int MAX_SMOKE = 100;
        // Add new Smoke-Particle whenever an old Smoke-Particel disappears
        // or at specified intervals.
        s->nextsmoke -= spf;
        if (s->nextsmoke <= 0 || removed_smoke) {
            s->nextsmoke = 0.12;
            Smoke* smoke = new Smoke();
            memcpy(smoke->pos, s->pos, sizeof(float)*3);
            smoke->cnt = 1.2f;
            smoke->timer = 0.0f;
            s->trail.push_back(smoke);
        } 

        // Now update the Missile itself
        //float tgt[] = { 0,0,0 };
        //if (s->target) memcpy(tgt, s->target->mPos, sizeof(float)*3);
        //float dir[] = { tgt[0]-s->pos[0], tgt[1]-s->pos[1], tgt[2]-s->pos[2] };
        //normalize(dir,dir);
        //normalize(s->vel,s->vel);
        //const float alpha = 0.96;
        //loop3i(s->vel[i] = (s->vel[i]*alpha+dir[i]*(1-alpha)));

        const float alpha = 0.992;
        loop3i(s->vel[i] *= alpha);
        
        const float speed = 5;
        s->pos[0] += speed * s->vel[0] * spf;
        s->pos[1] += speed * s->vel[1] * spf;
        s->pos[2] += speed * s->vel[2] * spf;
        s->fuel -= spf;
        s->timer += spf;

        // When fuel is empty just let the Missile and all
        // it's Smoke-Particles disappear.
        if (s->fuel < 0) {
            for (std::list<Smoke*>::iterator j = s->trail.begin(); j != s->trail.end();) {
                Smoke* smoke = *j++;
                s->trail.remove(smoke);
            }
            missiles.remove(s);
            delete s;
        } else {
            float v[] = {speed*s->vel[0], speed*s->vel[1], speed*s->vel[2] };
            int damaged = this->damageAllByCylinder(s->pos, v, spf, 5.0f, (1UL<<cObject::COLLIDEABLE)|(1UL<<cObject::DAMAGEABLE), 10);
            if (damaged) {
                missiles.remove(s);
                delete s;
            }
        }
    }
    
    mReloading -= spf;
    if (mReloading < 0) mReloading = 0.0f;
};


void cExplosion::drawSolid()
{
    glPushMatrix();
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);

    glColor4f(0.7,0.7,0.7,1.0);

    const float size = 0.8;
    const float length = size * 0.4;
    const float radius = size * 0.8;
    const float s1 = radius * sin(  0*PI_OVER_180);
    const float c1 = radius * cos(  0*PI_OVER_180);
    const float s2 = radius * sin(120*PI_OVER_180);
    const float c2 = radius * cos(120*PI_OVER_180);
    const float s3 = radius * sin(240*PI_OVER_180);
    const float c3 = radius * cos(240*PI_OVER_180);
    
    float m[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, m);
    
    for (std::list<Missile*>::iterator i = missiles.begin(); i != missiles.end();) {
        Missile* s = *i++;
        
        float r[3];
        cross(r, s->vel, (&m[4]));
        float u[3];
        cross(u, s->vel, r);
        float b[16];
        memset(b, 0, sizeof(float)*16);
        b[15] = 1;
        memcpy(&b[0], r, sizeof(float)*3);
        memcpy(&b[4], u, sizeof(float)*3);
        memcpy(&b[8], s->vel, sizeof(float)*3);
        //printMatrix(b);
        
        glPushMatrix();
            glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glMultMatrixf(b);
                //glNormAxis();
                glBegin(GL_TRIANGLE_FAN);
                glVertex3f(0,0,length);
                glVertex3f(s1,c1,0);
                glVertex3f(s2,c2,0);
                glVertex3f(s3,c3,0);
                glVertex3f(s1,c1,0);
                glEnd();
        glPopMatrix();
    }

    glPopMatrix();
    glPopAttrib();
}


void cExplosion::drawEffect()
{
    //return;
    
    glPushMatrix();
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glDepthMask(GL_FALSE);

    float n[16];
    glGetTransposeInverseRotationMatrix(n);
    
    // Draw Missile's flare.
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    for (std::list<Missile*>::iterator i = missiles.begin(); i != missiles.end();) {
        Missile* s = *i++;
        
        glPushMatrix();
            glTranslatef(s->pos[0], s->pos[1], s->pos[2]);            
                glMultMatrixf(n);
                float size = 2.5 + 2.5 * 0.01f*(rand()%100);
                glColor4f(0.8, 0.5, 0.1, 0.99f);
                glDisk(9+DETAIL, size*0.1f);
                size = 2.5 + 2.5 * 0.01f*(rand()%100);
                glColor4f(0.9, 0.9, 0.9, 0.99f);
                glDisk(7+DETAIL, size*0.07f);
        glPopMatrix();
   }
   
   // Draw Missile's trails.
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   for (std::list<Missile*>::iterator i = missiles.begin(); i != missiles.end();) {
        Missile* s = *i++;
        
        for (std::list<Smoke*>::iterator j = s->trail.begin(); j != s->trail.end();) {
            Smoke* smoke = *j++;
            glPushMatrix();
                glTranslatef(smoke->pos[0], smoke->pos[1], smoke->pos[2]);            
                    glMultMatrixf(n);
                    float m = 2*smoke->cnt;
                    glColor4f(m, m*0.8, m*0.6, 0.07+smoke->cnt);
                    float size = 0.1 + 0.3 * pow(1.0f+smoke->timer, 3.5);
                    glDisk(5+DETAIL, size);
            glPopMatrix();
        }
   }

    glPopMatrix();
    glPopAttrib();
}


